--
-- AutoCombine
-- Extended AICombine 
--
-- @author  Mogli aka biedens
-- @date  22.10.2013
--
--  code source: AICombine.lua by Giants Software    
 
AutoCombine = {};

function AutoCombine.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(Hirable, specializations) and SpecializationUtil.hasSpecialization(Steerable, specializations);
end;

local AcDirectory = g_currentModDirectory;

function AutoCombine:load(xmlFile)

	self.acCalibration = 1;
	self.acIsCalibrated = false;
	self.acNodeIsLinked = false;
	self.acButtons = {};

	self.acDelayTimeToStopMovement = 20;	
	self.acDelayTimeToMoveBack = 40;
	self.acDelayTimeToStopAutoCombine = 250;
	self.acOtherCombineActive = false;
	self.acWaitMode = false;

    self.startAutoCombine = SpecializationUtil.callSpecializationsFunction("startAutoCombine");
    self.stopAutoCombine = SpecializationUtil.callSpecializationsFunction("stopAutoCombine");   
	
	self.acOnTrafficCollisionTrigger = AutoCombine.acOnTrafficCollisionTrigger;
	self.acOnOtherCombineCollisionTrigger = AutoCombine.acOnOtherCombineCollisionTrigger;
	
	self.acI3D = getChild(loadI3DFile(Utils.getFilename("AutoCombine.i3d", AcDirectory)),"AutoCombine");
	
	self.acTrafficCollisionTrigger = getChild(self.acI3D,"vehicleCollisionTrigger");
	self.acBackTrafficCollisionTrigger = getChild(self.acI3D,"backCollisionTrigger");
	self.acOtherCombineCollisionTriggerL = getChild(self.acI3D,"otherCombColliTriggerL");
	self.acOtherCombineCollisionTriggerR = getChild(self.acI3D,"otherCombColliTriggerR");
	
	self.trafficCollisionIgnoreList = {};
    for k,v in pairs(self.components) do
        self.trafficCollisionIgnoreList[v.node] = true;
    end;
    self.acNumCollidingVehicles = 0;
    self.numToolsCollidingVehicles = {};
		
	self.onCombineTrailerTrigger = AutoCombine.onCombineTrailerTrigger;
    self.combineTrailerTriggers = {};  
    local i = 0;
	while true do
        local key = string.format("vehicle.combineTrailerTriggers.combineTrailerTrigger(%d)", i);
        if not hasXMLProperty(xmlFile, key) then
			break;
        end;
        local node = Utils.indexToObject(self.components, getXMLString(xmlFile,"vehicle.combineTrailerTriggers.combineTrailerTrigger#index"));
        if node ~= nil then
            local pipeState = Utils.getNoNil(getXMLInt(xmlFile, key.."#pipeState"), 2);
            self.combineTrailerTriggers[node] = {node=node, pipeState=pipeState};
        end;
        i = i + 1;
    end;
    for _, trigger in pairs(self.combineTrailerTriggers) do
        addTrigger(trigger.node, "onCombineTrailerTrigger", self);
    end;
  
    self.combineTrailersInRange = {};
    self.isTrailerInRange = false;
    self.trailerInRangePipeState = 0;
	self.waitingForTrailerToUnload = false;
    self.waitingForDischarge = false;
    self.waitForDischargeTime = 0;
    self.waitForDischargeTimeout = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.waitForDischargeTime"), 5000);
 
    self.isAutoCombineActivated = false;
    self.dtSum = 0;

    local aiMotorSound  = getXMLString(xmlFile, "vehicle.aiMotorSound#file");
    if aiMotorSound  ~= nil and aiMotorSound  ~= "" then
        aiMotorSound  = Utils.getFilename(aiMotorSound, self.baseDirectory);
        self.aiMotorSoundPitchOffset = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.motorSoundRun#pitchOffset"), 0);
        self.aiMotorSoundRadius = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.motorSoundRun#radius"), 50);
        self.aiMotorSoundInnerRadius = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.motorSoundRun#innerRadius"), 10);
        self.aiMotorSoundVolume = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.motorSoundRun#volume"), 1);
        self.aiMotorSound = createAudioSource("aiMotorSound", aiMotorSound, self.aiMotorSoundRadius, self.aiMotorSoundInnerRadius, self.aiMotorSoundVolume, 0);
        link(self.components[1].node, self.aiMotorSound);
        setVisibility(self.aiMotorSound, false);
    end;

    local translationPartNodeLeft = getChild(self.acI3D,"autoCombineAreaLeft");
    if translationPartNodeLeft ~= nil then
        self.acTranslationLeft = {};
        self.acTranslationLeft.node = translationPartNodeLeft;
        self.acTranslationLeft.minTrans = {};
        self.acTranslationLeft.minTrans[1] = 0;
        self.acTranslationLeft.minTrans[2] = 0;
        self.acTranslationLeft.minTrans[3] = 3.5;

        self.acTranslationLeft.maxTrans = {};
        self.acTranslationLeft.maxTrans[1] = 10;
        self.acTranslationLeft.maxTrans[2] = 0;
        self.acTranslationLeft.maxTrans[3] = 3.5;

        self.acTranslationLeft.transTime = 2500;
        self.acTranslationLeft.touchTransLimit = 10;
    end;
	
	local translationPartNodeRight = getChild(self.acI3D,"autoCombineAreaRight");
    if translationPartNodeRight ~= nil then
        self.acTranslationRight = {};
        self.acTranslationRight.node = translationPartNodeRight;
        self.acTranslationRight.minTrans = {};
        self.acTranslationRight.minTrans[1] = 0;
        self.acTranslationRight.minTrans[2] = 0;
        self.acTranslationRight.minTrans[3] = 3.5;

        self.acTranslationRight.maxTrans = {};
        self.acTranslationRight.maxTrans[1] = -10;
        self.acTranslationRight.maxTrans[2] = 0;
        self.acTranslationRight.maxTrans[3] = 3.5;

        self.acTranslationRight.transTime = 2500;
        self.acTranslationRight.touchTransLimit = 10;
    end;
	
	self.acTranslationRoot = {};
	self.acTranslationRoot.node = self.acI3D;
    self.acTranslationRoot.minTrans = {};
    self.acTranslationRoot.minTrans[1] = 0;
    self.acTranslationRoot.minTrans[2] = 0;
    self.acTranslationRoot.minTrans[3] = -5;

    self.acTranslationRoot.maxTrans = {};
    self.acTranslationRoot.maxTrans[1] = 0;
    self.acTranslationRoot.maxTrans[2] = 0;
    self.acTranslationRoot.maxTrans[3] = 5;

    self.acTranslationRoot.transTime = 2500;
    self.acTranslationRoot.touchTransLimit = 10;	

    self.acAreaLeft = {};
    self.acAreaLeft.startOutside  = getChild(translationPartNodeLeft,"areaStartOutside");
    self.acAreaLeft.widthOutside  = getChild(translationPartNodeLeft,"areaWidthOutside");
    self.acAreaLeft.heightOutside = getChild(translationPartNodeLeft,"areaHeightOutside");
    self.acAreaLeft.startInside   = getChild(translationPartNodeLeft,"areaStartInside");
    self.acAreaLeft.widthInside   = getChild(translationPartNodeLeft,"areaWidthInside");
    self.acAreaLeft.heightInside  = getChild(translationPartNodeLeft,"areaHeightInside");
	self.acAreaLeft.startCheck    = getChild(translationPartNodeLeft,"areaStartCheck");
	self.acAreaLeft.widthCheck    = getChild(translationPartNodeLeft,"areaWidthCheck");
	self.acAreaLeft.heightCheck   = getChild(translationPartNodeLeft,"areaHeightCheck");
    self.acAreaLeft.active = true;

    self.acAreaRight = {};
    self.acAreaRight.startOutside  = getChild(translationPartNodeRight,"areaStartOutside");
    self.acAreaRight.widthOutside  = getChild(translationPartNodeRight,"areaWidthOutside");
    self.acAreaRight.heightOutside = getChild(translationPartNodeRight,"areaHeightOutside");
    self.acAreaRight.startInside   = getChild(translationPartNodeRight,"areaStartInside");
    self.acAreaRight.widthInside   = getChild(translationPartNodeRight,"areaWidthInside");
	self.acAreaRight.heightInside  = getChild(translationPartNodeRight,"areaHeightInside");
	self.acAreaRight.startCheck    = getChild(translationPartNodeRight,"areaStartCheck");
	self.acAreaRight.widthCheck    = getChild(translationPartNodeRight,"areaWidthCheck");
	self.acAreaRight.heightCheck   = getChild(translationPartNodeRight,"areaHeightCheck");
    self.acAreaRight.active = false;

	self.acWorkWidth = nil;
    self.acEnabled = false;
	self.acDelayCounting = false;
    self.acDelay = 0;
	self.gotFruitArea = nil;	
    self.acAreaLeft.arrowLeft   = getChild(translationPartNodeLeft,"acWorkwidthArrow");
	self.acAreaRight.arrowRight = getChild(translationPartNodeRight,"acWorkwidthArrow");
	
	self.turnDirection = 0;
    self.autoRotateBackSpeedBackup = self.autoRotateBackSpeed;	

	self.acHelpPanelPath = Utils.getFilename("AutoCombineHud.dds", AcDirectory);
	self.acHudWidth = 0.19 --0.35;  
	self.acHudHeight = 0.3 ;  --0.377
	self.acHudPosX = 0.019;   
	self.acHudPosY = 0.0108;  
	self.acHelpPanelBtnPosX = 0.03;     
	self.acHelpPanelBtnPosY = 0.6;  --0.454 
	self.acHelpPanelTextPosX = self.acHelpPanelBtnPosX;--0.024;     
	self.acHelpPanelTextPosY = 0.02;  --0.454 
	self.acHelpPanelOverlay = Overlay:new("AutoCombineHud", self.acHelpPanelPath, self.acHudPosX, self.acHudPosY, self.acHudWidth, self.acHudHeight);
	self.helpPanelActive = false;	
	
	--self.acSetAreaLeft = AutoCombine.setAreaLeft;
	AutoCombine.addButton(self,Utils.getFilename("off.dds", AcDirectory), Utils.getFilename("on.dds", AcDirectory), AutoCombine.onStart, AutoCombine.evalStart, 0.00,0.32);
	AutoCombine.addButton(self,Utils.getFilename("inactive_left.dds", AcDirectory), Utils.getFilename("active_left.dds", AcDirectory), AutoCombine.setAreaLeft, AutoCombine.evalAreaLeft, 0.06,0.32);
	AutoCombine.addButton(self,Utils.getFilename("inactive_right.dds", AcDirectory), Utils.getFilename("active_right.dds", AcDirectory), AutoCombine.setAreaRight, AutoCombine.evalAreaRight, 0.12,0.32);
	
	AutoCombine.addButton(self,Utils.getFilename("no_wait.dds", AcDirectory),Utils.getFilename("wait.dds", AcDirectory), AutoCombine.setWait, AutoCombine.evalWait, 0.00,0.40);	
	AutoCombine.addButton(self,Utils.getFilename("bigger.dds", AcDirectory),nil, AutoCombine.setWidthUp, AutoCombine.evalWidthUp, 0.06,0.40);
	AutoCombine.addButton(self,Utils.getFilename("smaller.dds", AcDirectory),nil, AutoCombine.setWidthDown, AutoCombine.evalWidthDown, 0.12,0.40);

	AutoCombine.addButton(self,Utils.getFilename("no_distance.dds", AcDirectory),Utils.getFilename("distance.dds", AcDirectory), AutoCombine.setOtherCombine, AutoCombine.evalOtherCombine, 0.00,0.48);	
	AutoCombine.addButton(self,Utils.getFilename("forward.dds", AcDirectory), nil, AutoCombine.setForward, nil, 0.06,0.48);
	AutoCombine.addButton(self,Utils.getFilename("backward.dds", AcDirectory), nil, AutoCombine.setBackward, nil, 0.12,0.48);
	
	print("AutoCombine loaded");
end;

function AutoCombine.addButton(self, imgEnabled, imgDisabled, cbOnClick, cbVisible, x,y, width,height)
	local x1 = self.acHelpPanelBtnPosX + x;
	local y2 = self.acHelpPanelBtnPosY - y;
	local x2 = x1+0.045;
	if width ~= nil then
		local x2 = x1+abs(width);
	end;
	local y1 = y2-0.06;
	if width ~= nil then
		local y1 = y2-abs(height);
	end;
    local overlay = Overlay:new(nil, imgEnabled, x1,y1,x2-x1,y2-y1);
    local overlay2 = nil;
    if imgDisabled == nil then
		overlay2 = Overlay:new(nil, Utils.getFilename("empty.dds", AcDirectory), x1,y1,x2-x1,y2-y1);
	else
		overlay2 = Overlay:new(nil, imgDisabled, x1,y1,x2-x1,y2-y1);
    end;
    local button = {enabled=true, ovEnabled=overlay, ovDisabled=overlay2, onClick=cbOnClick, onVisible=cbVisible, twoState=(imgDisabled ~= nil), rect={x1,y1,x2,y2}};
    table.insert(self.acButtons, button);
    return button;
end;

function AutoCombine.renderButtons(self)
    --setTextAlignment(RenderText.ALIGN_LEFT);
    --setTextBold(false);

    for _,button in pairs(self.acButtons) do
        if button.onVisible ~= nil then
			button.enabled = button.onVisible(self);
		end;
		if button.enabled then
			if button.ovEnabled ~= nil then
				button.ovEnabled:render();
			end;
		else
			if button.ovDisabled ~= nil then
				button.ovDisabled:render();
			end;
		end;
    end;	
	--setTextBold(false);
end;

function AutoCombine.showGui(self,on)
	--if on then
		--g_gui.currentGui = self;
	--else
		--g_gui.currentGui = nil;
	--end;

    self.helpPanelActive = on;
    g_mouseControlsHelp.active = not on;
	InputBinding.setShowMouseCursor(on);		
end;

function AutoCombine:evalWidthUp()
	return self.acWorkWidth < 20;
end;

function AutoCombine:setWidthUp()
	self.translationBigger  = true;
	self.translationSmaller = not self.translationBigger;
end;

function AutoCombine:evalWidthDown()
	return self.acWorkWidth > 0;
end;

function AutoCombine:setWidthDown()
	self.translationBigger  = false;
	self.translationSmaller = not self.translationBigger;
end;

function AutoCombine:setForward()
	self.translationForward  = true;
	self.translationBackward = not self.translationForward ;
end;

function AutoCombine:setBackward()
	self.translationForward  = false;
	self.translationBackward = not self.translationForward ;
end;

function AutoCombine:evalWait()
	return not self.acWaitMode;
end;

function AutoCombine:setWait(enabled)
	self.acWaitMode = enabled;
end;

function AutoCombine:evalAreaLeft()
	return not self.acAreaLeft.active;
end;

function AutoCombine:setAreaLeft(enabled)
	if not enabled then return; end;
	self.acAreaLeft.active  = enabled;
	self.acAreaRight.active = not enabled;
	self.acDelay = 0;           
end;

function AutoCombine:evalAreaRight()
	return not self.acAreaRight.active;
end;

function AutoCombine:setAreaRight(enabled)
	if not enabled then return; end;
	self.acAreaRight.active = enabled;
	self.acAreaLeft.active  = not enabled;
	self.acDelay = 0;           
end;

function AutoCombine:evalStart()
	return not self.isAutoCombineActivated;
end;

function AutoCombine:onStart(enabled)
	if enabled == self.isAutoCombineActivated then
		return;
	end;
	
	if self.isAutoCombineActivated then
		self:stopAutoCombine();
		self.acEnabled = false;
		self.speed2Level = 0;
	else
		if    self.speed2Level ~= 1
		  and self.speed2Level ~= 2
		  and self.speed2Level ~= 3
		  and self.speed2Level ~= 4 then
			self.speed2Level = 1;
		end;
		self:startAutoCombine();
		self.acEnabled = true;        
	end;
	self.motor.speedLevel = 0;	
end;

function AutoCombine:evalOtherCombine()
	return not self.acOtherCombineActive;
end;

function AutoCombine:setOtherCombine(enabled)
	self.acOtherCombineActive = enabled;
	if not self.acOtherCombineActive then
		self.acNumCollidingVehicles = 0;
	end;	
end;

function AutoCombine:delete()

    self:stopAutoCombine();
	for _, combineTrailerTrigger in pairs(self.combineTrailerTriggers) do
        removeTrigger(combineTrailerTrigger.node);
    end;
	AutoCombine:removeOtherCombineCollisionTrigger()
	AutoCombine:removeBackTrafficCollisionTrigger();
end;

function AutoCombine:mouseEvent(posX, posY, isDown, isUp, button)
    if self.helpPanelActive and isDown and button == 1 then
        for _,overlayButton in pairs(self.acButtons) do
            if overlayButton.onClick ~= nil and overlayButton.rect[1] <= posX and posX <= overlayButton.rect[3] and overlayButton.rect[2] <= posY and posY <= overlayButton.rect[4] then
				if overlayButton.enabled then
					if overlayButton.twoState ~= nil then
						overlayButton.onClick(self, true);
					else
						overlayButton.onClick(self);
                    end;
				elseif overlayButton.twoState then
					overlayButton.onClick(self, false);
				end;
                break;				
            end;
        end;
	end;
end;

function AutoCombine:keyEvent(unicode, sym, modifier, isDown)
	
	if isDown and sym == Input.KEY_s then
		self.speed2Level = 0;
	end;	
end;

function AutoCombine:update(dt)

	--if self:getIsActiveForInput() then	
	if self:getIsActive() then	
		if InputBinding.hasEvent(InputBinding.AC_COMBINE_HELPPANEL) then
			AutoCombine.showGui( self, not self.helpPanelActive );
		end;
		if	InputBinding.hasEvent(InputBinding.SPEED_LEVEL1) then
			self.speed2Level = 1;
		elseif InputBinding.hasEvent(InputBinding.SPEED_LEVEL2) then
			self.speed2Level = 2;
		elseif InputBinding.hasEvent(InputBinding.SPEED_LEVEL3) then
			self.speed2Level = 3;
		elseif InputBinding.hasEvent(InputBinding.SPEED_LEVEL4) then
			self.speed2Level = 4;
		end;
		if InputBinding.hasEvent(InputBinding.AC_COMBINE_STARTSTOP) then
			AutoCombine.onStart( self,not self.isAutoCombineActivated );
		end;
--		if self.helpPanelActive then			
--			if	InputBinding.hasEvent(InputBinding.AC_COMBINE_AREALEFT) then
--				AutoCombine:setAreaLeft();   
--			end;
--			if	InputBinding.hasEvent(InputBinding.AC_COMBINE_AREARIGHT) then
--				AutoCombine:setAreaRight();   
--			end;
--			if	InputBinding.isPressed(InputBinding.AC_COMBINE_WORKWIDTHUP) then
--				self.translationBigger = true;
--			end;
--			if	InputBinding.isPressed(InputBinding.AC_COMBINE_WORKWIDTHDOWN) then
--				self.translationSmaller = true ;				
--			end;
--			if InputBinding.hasEvent(InputBinding.AC_COMBINE_STARTSTOP) then
--				if self.isAutoCombineActivated then
--					self:stopAutoCombine();
--					self.acEnabled = false;
--				else
--					if	self.motor.speedLevel == 0 then
--						self.speed2Level = 1;
--					end;
--					self:startAutoCombine();
--					self.acEnabled = true;        
--				end;
--			end;
--			if InputBinding.hasEvent(InputBinding.AC_COMBINE_WAITMODE_ONOFF) then 
--				self.acWaitMode = not self.acWaitMode 
--			end;
--			if InputBinding.hasEvent(InputBinding.AC_COMBINE_COLTRGGER_ONOFF) then 
--				self.acOtherCombineActive = not self.acOtherCombineActive
--				if not self.acOtherCombineActive then
--					self.acNumCollidingVehicles = 0;
--				end;	
--			end;			
--		end;
	end;
	
	if self.isAutoCombineActivated or self.helpPanelActive then
		if not self.acNodeIsLinked then
			self.acNodeIsLinked = true;
			link(self.components[1].node,self.acI3D);
		end;
	elseif self.acNodeIsLinked then
		self.acNodeIsLinked = false;
		unlink(self.acI3D);
	end;
end;

function AutoCombine:updateTick(dt)
	AutoCombine:removeOtherCombineCollisionTrigger();
	AutoCombine:removeBackTrafficCollisionTrigger();	
	if self.isServer then
		if self.isEntered then
			if self.AutoCombineDx ~= nil then
				local x, y, z = getTranslation(self.acTranslationRight.node);
				setTranslation(self.acTranslationRight.node, -self.AutoCombineDx,y,z);
				x, y, z = getTranslation(self.acTranslationLeft.node);
				setTranslation(self.acTranslationLeft.node, self.AutoCombineDx,y,z);
				print("self.AutoCombineDx "..self.AutoCombineDx);
				self.AutoCombineDx = nil;
			end;
			if self.AutoCombineDz ~= nil then
				local x, y, z = getTranslation(self.acTranslationRoot.node);
				setTranslation(self.acTranslationRoot.node, x,y,self.AutoCombineDz);
				print("self.AutoCombineDz "..self.AutoCombineDz);
				self.AutoCombineDz = nil;
			end;
			
			if self.acTranslationRight ~= nil and (self.translationBigger or self.translationSmaller) then
				local x, y, z = getTranslation(self.acTranslationRight.node);
				local trans = {x,y,z};
				local newTrans = Utils.getMovedLimitedValues(trans, self.acTranslationRight.maxTrans, self.acTranslationRight.minTrans, 3, self.acTranslationRight.transTime, dt, not self.translationBigger);
				setTranslation(self.acTranslationRight.node, unpack(newTrans));
			end;
			if self.acTranslationLeft ~= nil and (self.translationBigger or self.translationSmaller) then
				local x, y, z = getTranslation(self.acTranslationLeft.node);
				local trans = {x,y,z};
				local newTrans = Utils.getMovedLimitedValues(trans, self.acTranslationLeft.maxTrans, self.acTranslationLeft.minTrans, 3, self.acTranslationLeft.transTime, dt, not self.translationBigger);
				setTranslation(self.acTranslationLeft.node, unpack(newTrans));
			end;
			if self.acTranslationRoot ~= nil and (self.translationForward or self.translationBackward) then
				local x, y, z = getTranslation(self.acTranslationRoot.node);
				local trans = {x,y,z};
				local newTrans = Utils.getMovedLimitedValues(trans, self.acTranslationRoot.minTrans, self.acTranslationRoot.maxTrans, 3, self.acTranslationRoot.transTime, dt, not self.translationBackward);
				setTranslation(self.acTranslationRoot.node, unpack(newTrans));
			end;
			
			self.translationBigger   = false;
			self.translationSmaller  = false;
			self.translationForward  = false;
			self.translationBackward = false;
			
			if self.helpPanelActive then
				setVisibility(self.acAreaLeft.arrowLeft, true);
				setVisibility(self.acAreaRight.arrowRight, true);
			else
				setVisibility(self.acAreaLeft.arrowLeft, false);
				setVisibility(self.acAreaRight.arrowRight, false);
			end;
		else
			setVisibility(self.acAreaLeft.arrowLeft, false);
			setVisibility(self.acAreaRight.arrowRight, false);
		end;

		self.acWorkWidth = getTranslation(self.acTranslationLeft.node) * 2;		

		if self.isAutoCombineActivated then
			if self.isBroken then
				self:stopAutoCombine();
			end;		
			self.dtSum = self.dtSum + dt;
			if self.dtSum > 20 then
				AutoCombine.updateAIMovement(self, self.dtSum);
				self.dtSum = 0;
			end;
			if (self.grainTankFillLevel > 0 or self.grainTankCapacity <= 0) and (self.grainTankFillLevel >= self.grainTankCapacity or self.isTrailerInRange) then
                if self.trailerInRangePipeState > 0 then
                    self:setPipeState(self.trailerInRangePipeState);
					if self.acWaitMode then
						self.waitingForDischarge = true;
					end;
                else
                    self:setPipeState(2);
                end;
                --self:setPipeOpening(true);
                if self.isTrailerInRange then
                    self.waitForDischargeTime = self.time + self.waitForDischargeTimeout;
                end;
                if self.acWaitMode then 
					if self.grainTankFillLevel >= self.grainTankCapacity and self.grainTankCapacity > 0 then                    
						self.waitingForDischarge = true;
						self.waitForDischargeTime = self.time + self.waitForDischargeTimeout;
					end;
				else
					if self.grainTankFillLevel < self.grainTankCapacity and self.grainTankCapacity > 0 then                    
						self.waitingForDischarge = false;
						self.waitForDischargeTime = self.time + self.waitForDischargeTimeout;
						if self:getIsThreshingAllowed(true) then
							self:setIsThreshing(true);
						end; 
					end;
				end;
            else
                -- no trailer in range and not full
					if (self.waitingForDischarge and self.grainTankFillLevel <= 0) or self.waitForDischargeTime <= self.time then
						self.waitingForDischarge = false;
						if not self.isTrailerInRange then
							-- only close the pipe if no trailer is in range
							self:setPipeState(1);
						end;
						--self:setPipeOpening(false);
						if self:getIsThreshingAllowed(true) then
							self:setIsThreshing(true);
						end;                
					end;
            end;			
			AutoCombine:addOtherCombineCollisionTrigger()
		else
			self.dtSum = 0;
			AutoCombine:removeOtherCombineCollisionTrigger()
			AutoCombine:removeBackTrafficCollisionTrigger();
		end;		
	end;
		
end;

function AutoCombine:startAutoCombine()
    self:hire();
    if not self.isAutoCombineActivated then
		self.isAutoCombineActivated = true;		
		--self.isTrailerInRange = false;
		self.waitingForDischarge = false;
		if not self.acAreaLeft.active and not self.acAreaRight.active then
			self.acAreaLeft.active = true;
		end;
		if self.acEnabled then
			self.autoRotateBackSpeed = 0;
		else
			self.autoRotateBackSpeed = self.autoRotateBackSpeedBackup;
		end;
        self.acNumCollidingVehicles = 0;
		if self.acTrafficCollisionTrigger ~= nil then
            addTrigger(self.acTrafficCollisionTrigger, "acOnTrafficCollisionTrigger", self);
        end;  		
        self.checkSpeedLimit = false;
        setVisibility(self.aiMotorSound, true);
		self.acCalibration = 1;
		self.acIsCalibrated = false;
    end;
end;

function AutoCombine:stopAutoCombine()
    self:dismiss();
    if self.isAutoCombineActivated then
        self.isAutoCombineActivated = false;
		self.motor:setSpeedLevel(0, false);
        self.motor.maxRpmOverride = nil;
        WheelsUtil.updateWheelsPhysics(self, 0, self.lastSpeed, 0, false, self.requiredDriveMode);	
		if self.acTrafficCollisionTrigger ~= nil then
			removeTrigger(self.acTrafficCollisionTrigger, "acOnTrafficCollisionTrigger", self);
		end;         
        self.checkSpeedLimit = true;
        setVisibility(self.aiMotorSound, false);			
		self.acDelay = 0;
		self.acDelay = 0;
		self:setIsThreshing(false);		
	end;
end;

function AutoCombine.updateAIMovement(self, dt, onLeave)

    local allowedToDrive = true;
	
--use Lights when needed_begin	
	if not self.isControlled then
        if g_currentMission.environment.needsLights then
            self:setLightsVisibility(true);
        else
            self:setLightsVisibility(false);
        end;
    end;
--use Lights when needed_end
	
	if self.grainTankCapacity == 0 then
		if not self.pipeStateIsUnloading[self.currentPipeState] then
			allowedToDrive = false;
		end;
		if not self.isPipeUnloading and (self.lastArea > 0 or self.lastLostGrainTankFillLevel > 0) then
        -- there is some fruit to unload, but there is no trailer. Stop and wait for a trailer
			self.waitingForTrailerToUnload = true;
		end;
    else
		if self.grainTankFillLevel >= self.grainTankCapacity then
			allowedToDrive = false;
        end
    end	
	
    if self.waitingForTrailerToUnload then
        if self.lastValidGrainTankFruitType ~= FruitUtil.FRUITTYPE_UNKNOWN then
            local trailer = self:findTrailerToUnload(self.lastValidGrainTankFruitType);
            if trailer ~= nil then
               -- there is a trailer to unload. Continue working
                self.waitingForTrailerToUnload = false;
            end;
        else
           -- we did not cut anything yet. We shouldn't have ended in this state. Just continue working
            self.waitingForTrailerToUnload = false;
        end;
    end;
	
    if (self.grainTankFillLevel >= self.grainTankCapacity and self.grainTankCapacity > 0) or self.waitingForTrailerToUnload or self.waitingForDischarge then
        allowedToDrive = false;
    end;
	
	if not self:getIsThreshingAllowed(true) then
        allowedToDrive = false;
        self:setIsThreshing(false);
        self.waitingForWeather = true;
    elseif self.waitingForWeather then
        self:startThreshing();
        self.waitingForWeather = false;
    end;
	
	if self.acNumCollidingVehicles > 0 then
        allowedToDrive = false;
    end;
    for k,v in pairs(self.numToolsCollidingVehicles) do
        if v > 0 then
            allowedToDrive = false;
            break;
        end;
    end;

	if self.acEnabled then --and self.attachedCutter.hasGroundContact then | and self.movingDirection > 0
									
		self.turnDirection = 0;

		if self.lastValidInputFruitType == FruitUtil.FRUITTYPE_UNKNOWN then
			moveForwards = true;
			self.acDelayCounting = true;
			for cutter,implement in pairs(self.attachedCutters) do
				local jointDesc = self.attacherJoints[implement.jointDescIndex];
				jointDesc.moveDown = true;
			end;						
	    else
			local currentNode = nil;
			if     self.acAreaRight.active then
				currentNode = self.acAreaRight;
				otherNode   = self.acAreaLeft;
			elseif self.acAreaLeft.active then
				currentNode = self.acAreaLeft;
				otherNode   = self.acAreaRight;
			end;
				
			if currentNode ~= nil and otherNode ~= nil then
                local x,y,z = getWorldTranslation(currentNode.startOutside);
                local x1,y1,z1 = getWorldTranslation(currentNode.widthOutside);
                local x2,y2,z2 = getWorldTranslation(currentNode.heightOutside);
                local outside = Utils.getFruitArea(self.lastValidInputFruitType, x, z, x1, z1, x2, z2);

                local x,y,z = getWorldTranslation(currentNode.startInside);
                local x1,y1,z1 = getWorldTranslation(currentNode.widthInside);
                local x2,y2,z2 = getWorldTranslation(currentNode.heightInside);
                local inside = Utils.getFruitArea(self.lastValidInputFruitType, x, z, x1, z1, x2, z2);
                
				local x,y,z = getWorldTranslation(currentNode.startCheck);
                local x1,y1,z1 = getWorldTranslation(currentNode.widthCheck);
                local x2,y2,z2 = getWorldTranslation(currentNode.heightCheck);
                local outsideCheck = Utils.getFruitArea(self.lastValidInputFruitType, x, z, x1, z1, x2, z2);                    
				
				local x,y,z = getWorldTranslation(otherNode.startOutside);
                local x1,y1,z1 = getWorldTranslation(otherNode.widthOutside);
                local x2,y2,z2 = getWorldTranslation(otherNode.heightOutside);
                local insideCheck = Utils.getFruitArea(self.lastValidInputFruitType, x, z, x1, z1, x2, z2);	

				if 0 < inside and ( inside < self.acCalibration or not self.acIsCalibrated ) then
					self.acCalibration  = inside;
					self.acIsCalibrated = true;
				end;
				if 0 < outside and ( outside < self.acCalibration or not self.acIsCalibrated ) then
					self.acCalibration  = outside;
					self.acIsCalibrated = true;
				end;					
				
				inside  = inside / self.acCalibration;
				outside = outside / self.acCalibration;
				
				--self.turnDirection = outside -(9 -inside);
                self.turnDirection = outside*4 -(20 -inside);

                if (inside < 1 and outside < 1) then
                    self.acDelayCounting = true;
					self.gotFruitArea = false;
                else
					self.acDelayCounting = false;
                    self.acDelay = 0;
					self.gotFruitArea = true;
                end;			
				if self.acDelay >= self.acDelayTimeToStopMovement then
					allowedToDrive = false;						
					for cutter,implement in pairs(self.attachedCutters) do
						local jointDesc = self.attacherJoints[implement.jointDescIndex];
						jointDesc.moveDown = false;
					end;						
				end;
				if self.acDelay >= self.acDelayTimeToMoveBack then
					if self.acNumCollidingVehicles > 0 then
						allowedToDrive = false;
						self.acDelayCounting = false;
					else
						allowedToDrive = true;
						self.acDelayCounting = true;
					end;
					moveForwards = false;
					self.turnDirection = ((20 -outside) -inside*4);
				end;
				if not self.gotFruitArea and outsideCheck >0 then
					moveForwards = true;
					--self.turnDirection = 0;
					self.acDelayCounting = false;
					self.acDelay = 0;
					for cutter,implement in pairs(self.attachedCutters) do
					local jointDesc = self.attacherJoints[implement.jointDescIndex];
						jointDesc.moveDown = true;
					end;
					if insideCheck <= 0 then
						self.turnDirection = 0;
					--else
						--self.turnDirection = outside -(9 -inside);
						--self.turnDirection = outside*4 -(20 -inside);
					end;
				elseif self.gotFruitArea then
					moveForwards = true;
					--self.turnDirection = outside -(9 -inside);
					self.turnDirection = outside*4 -(20 -inside);
					for cutter,implement in pairs(self.attachedCutters) do
						local jointDesc = self.attacherJoints[implement.jointDescIndex];
						jointDesc.moveDown = true;
					end;						
				end;                
			end;
			
		end;
		
		if self.acAreaLeft.active then
			self.turnDirection = - self.turnDirection;
		end;
			
		local acceleration = 0;					
		if self.isMotorStarted then
			if self.motor.speedLevel ~= 0 then
				acceleration = 1.0;
			end;
		end;
		if self.fuelFillLevel == 0 then
			acceleration = 0;
		end;
			
		if self.acDelayCounting then
			self.acDelay = self.acDelay +1;
		end;

		--if self.turnDirection < -19 then self.turnDirection = -19; end;
		--if self.turnDirection >  19 then self.turnDirection =  19; end;
				
		local rotScale = math.min(1.0/(self.lastSpeed*50+1), 1);
		local factor   = math.abs(self.turnDirection);
		local f_min    = 0.0001*(factor+4)^3;
		local f_factor = 10^(3*(factor^0.39-2.5));
				
		if self.turnDirection < -19 then
            self.rotatedTime = math.min(self.rotatedTime - dt/1000*self.turnDirection*rotScale, 0.90);-----self.minRotTime
		elseif self.turnDirection > 19 then
            self.rotatedTime = math.max(self.rotatedTime - dt/1000*self.turnDirection*rotScale, -0.90);-----self.minRotTime
		elseif self.turnDirection < -1 then
            self.rotatedTime = math.min(self.rotatedTime - dt/1000*self.turnDirection*f_factor, f_min);----self.maxRotTime
		elseif self.turnDirection > 1 then
            self.rotatedTime = math.max(self.rotatedTime - dt/1000*self.turnDirection*f_factor, -f_min);-----self.minRotTime
        else
            self.rotatedTime = 0;
        end;
				
		if self.firstTimeRun then
			local acc = acceleration;
			if speedLevel ~= nil and speedLevel ~= 0 then
				acc = self.motor.accelerations[speedLevel];
				self.motor:setSpeedLevel(speedLevel, true);
				if math.abs(angle) >= slowAngleLimit then
					self.motor.maxRpmOverride = self.motor.maxRpm[speedLevel]*slowMaxRpmFactor;
				else
					self.motor.maxRpmOverride = nil;
				end;
			end;
			if not allowedToDrive then
				acc = 0;
			end;
			if not moveForwards then
				acc = -acc;
				AutoCombine:addBackTrafficCollisionTrigger()
			else
				AutoCombine:removeBackTrafficCollisionTrigger()
			end;
			if self.maxAccelerationSpeed ~= nil then
				acc = Steerable.calculateRealAcceleration(self, acc, dt);
			end;					
			WheelsUtil.updateWheelsPhysics(self, dt, self.lastSpeed, acc, not allowedToDrive, self.requiredDriveMode);
		end;
			
		if self.steering ~= nil then
			setRotation(self.steering, 0, self.rotatedTime*self.steeringSpeed, 0);
		end;
				
		if self.acDelay > self.acDelayTimeToStopAutoCombine then
			self:stopAutoCombine();		
		end;					
					
		if allowedToDrive then				
			self.motor.speedLevel = self.speed2Level;		
		else
			self.motor.speedLevel = 0;
		end;
									
	else
		self.acDelay = 0;
	end;	
end;

function AutoCombine:addBackTrafficCollisionTrigger()
	
	if self.acBackTrafficCollisionTrigger ~= nil then
		addTrigger(self.acBackTrafficCollisionTrigger, "acOnTrafficCollisionTrigger", self);
	end;
end;
function AutoCombine:removeBackTrafficCollisionTrigger()
	
	if self.acBackTrafficCollisionTrigger ~= nil then		
		removeTrigger(self.acBackTrafficCollisionTrigger);
	end;
end;

function AutoCombine:addOtherCombineCollisionTrigger()

	if self.acOtherCombineActive then	
		if self.acOtherCombineCollisionTriggerL ~= nil and self.acOtherCombineCollisionTriggerR ~= nil then
			if self.acAreaLeft.active then
				addTrigger(self.acOtherCombineCollisionTriggerL, "acOnOtherCombineCollisionTrigger", self);
				removeTrigger(self.acOtherCombineCollisionTriggerR, "acOnOtherCombineCollisionTrigger", self);
				
			elseif self.acAreaRight.active then
				addTrigger(self.acOtherCombineCollisionTriggerR, "acOnOtherCombineCollisionTrigger", self);
				removeTrigger(self.acOtherCombineCollisionTriggerL, "acOnOtherCombineCollisionTrigger", self);
			end;
		end;
	end;	
end;
function AutoCombine:removeOtherCombineCollisionTrigger()

	if self.acOtherCombineCollisionTriggerL ~= nil and self.acOtherCombineCollisionTriggerR ~= nil then
		removeTrigger(self.acOtherCombineCollisionTriggerL);		
		removeTrigger(self.acOtherCombineCollisionTriggerR);	
    end;
end;
		
function AutoCombine:acOnTrafficCollisionTrigger(triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)
    if onEnter or onLeave then
        if otherId == g_currentMission.player.rootNode then
            if onEnter then
                self.acNumCollidingVehicles = self.acNumCollidingVehicles+1;
            elseif onLeave then
                self.acNumCollidingVehicles = math.max(self.acNumCollidingVehicles-1, 0);
            end;
        else
            local vehicle = g_currentMission.nodeToVehicle[otherId];
            if vehicle ~= nil and self.trafficCollisionIgnoreList[otherId] == nil then
                if onEnter then
                    self.acNumCollidingVehicles = self.acNumCollidingVehicles+1;
                elseif onLeave then
                    self.acNumCollidingVehicles = math.max(self.acNumCollidingVehicles-1, 0);
                end;
            end;
        end;
    end;
end;

function AutoCombine:acOnOtherCombineCollisionTrigger(triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)
    if onEnter or onLeave then
        if otherId == Player.rootNode then
            if onEnter then
                self.acNumCollidingVehicles = self.acNumCollidingVehicles+1;
            elseif onLeave then
                self.acNumCollidingVehicles = math.max(self.acNumCollidingVehicles-1, 0);
            end;
        else            
			local cutter = g_currentMission.nodeToVehicle[otherId];
            if cutter ~= nil and self.trafficCollisionIgnoreList[otherId] == nil then
                if onEnter then
                    self.acNumCollidingVehicles = self.acNumCollidingVehicles+1;
                elseif onLeave then
                    self.acNumCollidingVehicles = math.max(self.acNumCollidingVehicles-1, 0);
                end;
            end;
        end;
    end;
end;

function AutoCombine:onCombineTrailerTrigger(triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)
    if onEnter or onLeave then
		local trailer = g_currentMission.nodeToVehicle[otherId];
        if trailer ~= nil and trailer.fillRootNode ~= nil then
            if onEnter then
                self.combineTrailersInRange[trailer] = self.combineTrailerTriggers[triggerId].pipeState;
                self.trailerInRangePipeState = math.max(self.trailerInRangePipeState, self.combineTrailerTriggers[triggerId].pipeState);
                self.isTrailerInRange = true;				
            else
                self.combineTrailersInRange[trailer] = nil;
                self.isTrailerInRange = false;
                self.trailerInRangePipeState = 0;
                for trailer, pipeState in pairs(self.combineTrailersInRange) do
                    self.trailerInRangePipeState = math.max(self.trailerInRangePipeState, pipeState);
                    self.isTrailerInRange = true;
                end;
            end;
        end;
    end;
end;

function AutoCombine:draw()

	if self.helpPanelActive then
		setTextAlignment(RenderText.ALIGN_LEFT);
		
		g_currentMission:addHelpButtonText(Steerable.AC_COMBINE_TEXTHELPPANELOFF, InputBinding.AC_COMBINE_HELPPANEL);
		self.acHelpPanelOverlay:render();
--		setTextColor(1.0, 1.0, 1.0, 1.0);	
--		if self.isAutoCombineActivated then
--		    renderText(self.acHelpPanelTextPosX, self.acHelpPanelTextPosY-0.225, 0.021,string.format(Steerable.AC_COMBINE_TXT_STOP));
--		else
--		    renderText(self.acHelpPanelTextPosX, self.acHelpPanelTextPosY-0.225, 0.021,string.format(Steerable.AC_COMBINE_TXT_START));
--		end;
--		--renderText(self.acHelpPanelTextPosX, self.acHelpPanelTextPosY-0.2725, 0.021, InputBinding.getKeyNamesOfDigitalAction(InputBinding.AC_COMBINE_STARTSTOP));--
--
		renderText(self.acHelpPanelTextPosX, self.acHelpPanelTextPosY, 0.021,string.format(Steerable.AC_COMBINE_TXT_WORKWIDTH.." %0.1fm",self.acWorkWidth));
--				
--		setTextColor(1.0, 1.0, 1.0, 1.0);			
--		
--		if self.acAreaLeft.active then
--			renderText(self.acHelpPanelTextPosX, self.acHelpPanelTextPosY-0.32, 0.021, string.format(Steerable.AC_COMBINE_TXT_ACTIVESIDELEFT));
--		else
--		    renderText(self.acHelpPanelTextPosX, self.acHelpPanelTextPosY-0.32, 0.021, string.format(Steerable.AC_COMBINE_TXT_ACTIVESIDERIGHT));
--		end;	
--		
--		if self.acWaitMode then
--			renderText(self.acHelpPanelTextPosX, self.acHelpPanelTextPosY-0.405, 0.021, string.format(Steerable.AC_COMBINE_WAITMODE_ON));
--		else
--		    renderText(self.acHelpPanelTextPosX, self.acHelpPanelTextPosY-0.405, 0.021, string.format(Steerable.AC_COMBINE_WAITMODE_OFF));
--		end;	
--		
--		if self.acOtherCombineActive then 
--			renderText(self.acHelpPanelTextPosX, self.acHelpPanelTextPosY-0.515, 0.019,string.format(Steerable.AC_COMBINE_COLLISIONTRIGGERMODE_ON));  
--		else 
--			renderText(self.acHelpPanelTextPosX, self.acHelpPanelTextPosY-0.515, 0.019,string.format(Steerable.AC_COMBINE_COLLISIONTRIGGERMODE_OFF));
--		end;
		
		AutoCombine.renderButtons(self);
	else
		g_currentMission:addHelpButtonText(Steerable.AC_COMBINE_TEXTHELPPANELON, InputBinding.AC_COMBINE_HELPPANEL);
	end;

	if self.isAutoCombineActivated then
		g_currentMission:addHelpButtonText(Steerable.AC_COMBINE_TXT_STOP, InputBinding.AC_COMBINE_STARTSTOP);
	elseif self.helpPanelActive then
		g_currentMission:addHelpButtonText(Steerable.AC_COMBINE_TXT_START, InputBinding.AC_COMBINE_STARTSTOP);
	end;
end;

function acBool2int(boolean)
	if boolean then
		return 1;
	end;
	return 0;
end;

function AutoCombine:onLeave()
    g_mouseControlsHelp.active = true;
	InputBinding.setShowMouseCursor(false);		
end;

function AutoCombine:onEnter()
    g_mouseControlsHelp.active = not self.helpPanelActive;
	InputBinding.setShowMouseCursor(self.helpPanelActive);		
end;

function AutoCombine:getSaveAttributesAndNodes(nodeIdent)
	local attributes = "";
	local x,y,z1 = getTranslation(self.acTranslationLeft.node);
	local x1,y,z = getTranslation(self.acTranslationRoot.node);
	if   self.acWaitMode 
	  or self.acAreaRight.active
	  or self.acOtherCombineActive
	  or x > 0
	  or z ~= 0 then
		attributes = 'acWaitMode="'..acBool2int(self.acWaitMode).. '" ';
		attributes = attributes..'acAreaRight="'..acBool2int(self.acAreaRight.active).. '" ';
		attributes = attributes..'acOtherCombine="'..acBool2int(self.acOtherCombineActive).. '" ';
		attributes = attributes..'AutoCombineDx="'..x..'" ';		
		attributes = attributes..'AutoCombineDz="'..z..'" ';
	end;
	
	print(attributes);
	
	return attributes
end;

function AutoCombine:loadFromAttributesAndNodes(xmlFile, key, resetVehicles)
	local l = nil;
	l = getXMLInt(xmlFile, key.."#acWaitMode");
	if l~= nil then
		self.acWaitMode = (l == 1);
	end;
	l = getXMLInt(xmlFile, key.."#acAreaRight");
	if l~= nil then
		self.acAreaRight.active = (l == 1);
		self.acAreaLeft.active  = not self.acAreaRight.active;
		self.acDelay = 0;           
	end;
	l = getXMLInt(xmlFile, key.."#acOtherCombine");
	if l~= nil then
		self.acOtherCombineActive = enabled;
		if not self.acOtherCombineActive then
			self.acNumCollidingVehicles = 0;
		end;	
	end;	

	self.AutoCombineDx = nil;
	self.AutoCombineDy = nil;
	local x = getXMLFloat(xmlFile, key.."#AutoCombineDx"); if x ~= nil then self.AutoCombineDx = x; end;
	local z = getXMLFloat(xmlFile, key.."#AutoCombineDz"); if z ~= nil then self.AutoCombineDz = z; end;

	return BaseMission.VEHICLE_LOAD_OK;
end

